# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six

from os_acc import exception
from os_acc.tests.unit import base


class ACCExceptionTestCase(base.TestCase):
    def test_default_error_msg(self):
        class FakeACCException(exception.ExceptionBase):
            msg_fmt = "default message"

        exc = FakeACCException()
        self.assertEqual(six.text_type(exc), 'default message')

    def test_error_msg(self):
        self.assertEqual(six.text_type(exception.ExceptionBase('test')),
                         'test')

    def test_default_error_msg_with_kwargs(self):
        class FakeACCException(exception.ExceptionBase):
            msg_fmt = "default message: %(foo)s"

        exc = FakeACCException(foo="bar")
        self.assertEqual(six.text_type(exc), 'default message: bar')



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

import os_acc
from os_acc import pci
from os_acc.tests.unit import base


class TestPCI(base.TestCase):

    def setUp(self):
        super(TestPCI, self).setUp()

    @mock.patch.object(pci.PciAccelerator, 'attach')
    def attach_pci_device(self, mock_attach):
        os_acc.initialize(reset=True)
        fake_acc = {"instance_uuid": "", "type": "pci",
                    "name": "NVIDIA Corporation GP100GL",
                    "pcie_address": "0000:81:00.0"}

        class FakeGuest(object):
            def attach_device(self, conf, persistent, live):
                pass

            def detach_device(self, conf, persistent, live):
                pass

            def get_power_state(self, host):
                pass

        fake_instance = {}
        fake_inst_type = {}
        fake_guest = FakeGuest()
        fake_host = ""
        os_acc.attach(fake_acc, fake_guest, fake_host, fake_instance,
                      fake_inst_type)
        mock_attach.assert_called_once_with(fake_acc, fake_guest,
                                            fake_host, fake_instance,
                                            fake_inst_type)

    @mock.patch.object(pci.PciAccelerator, 'detach')
    def detach_pci_device(self, mock_detach):
        os_acc.initialize(reset=True)
        fake_acc = {"instance_uuid": "", "type": "pci",
                    "name": "NVIDIA Corporation GP100GL",
                    "pcie_address": "0000:81:00.0"}

        class FakeGuest(object):
            def attach_device(self, conf, persistent, live):
                pass

            def detach_device(self, conf, persistent, live):
                pass

            def get_power_state(self, host):
                pass

        fake_instance = {}
        fake_inst_type = {}
        fake_guest = FakeGuest()
        fake_host = ""
        os_acc.detach(fake_acc, fake_guest, fake_host, fake_instance,
                      fake_inst_type)
        mock_detach.assert_called_once_with(fake_acc, fake_guest,
                                            fake_host, fake_instance,
                                            fake_inst_type)



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from os_acc.i18n import _


class ExceptionBase(Exception):
    """Base Exception

    To correctly use this class, inherit from it and define
    a 'msg_fmt' property. That msg_fmt will get printf'd
    with the keyword arguments provided to the constructor.

    """
    msg_fmt = _("An unknown exception occurred.")

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if not message:
            try:
                message = self.msg_fmt % kwargs
            except Exception:
                # at least get the core message out if something happened
                message = self.msg_fmt

        self.message = message
        super(ExceptionBase, self).__init__(message)

    def format_message(self):
        return self.args[0]


class LibraryNotInitialized(ExceptionBase):
    msg_fmt = _("Before using the os_acc library, you need to call "
                "os_acc.initialize()")


class AttachException(ExceptionBase):
    msg_fmt = _("Failed to attach ACC %(acc)s. Got error: %(err)s")


class DetachException(ExceptionBase):
    msg_fmt = _("Failed to detach ACC %(acc)s. Got error: %(err)s")


class InternalError(ExceptionBase):
    msg_fmt = _("%(err)s")


class NoMatchingPlugin(ExceptionBase):
    msg_fmt = _("No ACC plugin was found with the name %(plugin_name)s")



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_config import cfg
from stevedore import extension

import os_acc
from os_acc import base as plugin_base
from os_acc import exception
from os_acc.tests.unit import base


class DemoPlugin(plugin_base.PluginBase):

    CONFIG_OPTS = (
        cfg.BoolOpt("make_it_work",
                    default=False,
                    help="Make everything work correctly by setting this"),
        cfg.IntOpt("sleep_time",
                   default=0,
                   help="How long to artifically sleep")
    )

    def get_config(self, acc, instance=None, inst_type=None):
        pass

    def attach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        pass

    def detach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        pass


class DemoPluginNoConfig(plugin_base.PluginBase):

    def get_config(self, acc, instance=None, inst_type=None):
        pass

    def attach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        pass

    def detach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        pass


class TestOSACC(base.TestCase):

    def setUp(self):
        super(TestOSACC, self).setUp()
        os_acc._EXT_MANAGER = None

    @mock.patch('stevedore.extension.ExtensionManager')
    def test_initialize(self, mock_EM):
        self.assertIsNone(os_acc._EXT_MANAGER)
        # Note: the duplicate call for initialize is to validate
        # that the extension manager is only initialized once
        os_acc.initialize()
        os_acc.initialize()
        mock_EM.assert_called_once_with(
            invoke_on_load=False, namespace='os_acc')
        self.assertIsNotNone(os_acc._EXT_MANAGER)

    def test_load_plugin(self):
        obj = DemoPlugin.load("demo")
        self.assertTrue(hasattr(cfg.CONF, "os_acc_demo"))
        self.assertTrue(hasattr(cfg.CONF.os_acc_demo, "make_it_work"))
        self.assertTrue(hasattr(cfg.CONF.os_acc_demo, "sleep_time"))
        self.assertEqual(cfg.CONF.os_acc_demo.make_it_work, False)
        self.assertEqual(cfg.CONF.os_acc_demo.sleep_time, 0)

        self.assertEqual(obj.config, cfg.CONF.os_acc_demo)

    def test_load_plugin_no_config(self):
        obj = DemoPluginNoConfig.load("demonocfg")
        self.assertFalse(hasattr(cfg.CONF, "os_acc_demonocfg"))

        self.assertIsNone(obj.config)

    def test_attach_not_initialized(self):
        self.assertRaises(
            exception.LibraryNotInitialized,
            os_acc.attach, None, None, None, None, None)

    def test_detach_not_initialized(self):
        self.assertRaises(
            exception.LibraryNotInitialized,
            os_acc.detach, None, None, None, None, None)

    @mock.patch.object(DemoPlugin, "attach")
    def test_attach(self, mock_attach):
        plg = extension.Extension(name="demo",
                                  entry_point="os-acc",
                                  plugin=DemoPlugin,
                                  obj=None)
        with mock.patch('stevedore.extension.ExtensionManager.names',
                        return_value=['foobar']),\
                mock.patch('stevedore.extension.ExtensionManager.__getitem__',
                           return_value=plg):
            os_acc.initialize()
            fake_acc = {"instance_uuid": "", "type": "demo",
                        "name": "NVIDIA Corporation GP100GL",
                        "pcie_address": "0000:81:00.0"}
            fake_instance = {}
            fake_inst_type = {}
            fake_guest = ""
            fake_host = ""
            os_acc.attach(fake_acc, fake_guest, fake_host, fake_instance,
                          fake_inst_type)
            mock_attach.assert_called_once_with(fake_acc, fake_guest,
                                                fake_host, fake_instance,
                                                fake_inst_type)

    @mock.patch.object(DemoPlugin, "get_config")
    def test_get_config(self, mock_get_config):
        plg = extension.Extension(name="demo",
                                  entry_point="os-acc",
                                  plugin=DemoPlugin,
                                  obj=None)
        with mock.patch('stevedore.extension.ExtensionManager.names',
                        return_value=['foobar']),\
                mock.patch('stevedore.extension.ExtensionManager.__getitem__',
                           return_value=plg):
            os_acc.initialize()
            fake_acc = {"instance_uuid": "", "type": "demo",
                        "name": "NVIDIA Corporation GP100GL",
                        "pcie_address": "0000:81:00.0"}
            fake_instance = {}
            fake_inst_type = {}
            os_acc.get_config(fake_acc, fake_instance, fake_inst_type)
            mock_get_config.assert_called_once_with(fake_acc, fake_instance,
                                                    fake_inst_type)

    @mock.patch.object(DemoPlugin, "detach")
    def test_detach(self, mock_detach):
        plg = extension.Extension(name="demo",
                                  entry_point="os-acc",
                                  plugin=DemoPlugin,
                                  obj=None)
        with mock.patch('stevedore.extension.ExtensionManager.names',
                        return_value=['foobar']),\
                mock.patch('stevedore.extension.ExtensionManager.__getitem__',
                           return_value=plg):
            os_acc.initialize()
            fake_acc = {"instance_uuid": "", "type": "demo",
                        "name": "NVIDIA Corporation GP100GL",
                        "pcie_address": "0000:81:00.0"}
            fake_instance = {}
            fake_inst_type = {}
            fake_guest = ""
            fake_host = ""
            os_acc.detach(fake_acc, fake_guest, fake_host, fake_instance,
                          fake_inst_type)
            mock_detach.assert_called_once_with(fake_acc, fake_guest,
                                                fake_host, fake_instance,
                                                fake_inst_type)



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from os_acc import base


class PciAccelerator(base.PluginBase):

    def get_config(self, acc, instance=None, inst_type=None):
        pcie_addr = acc["pcie_address"]
        domain, bus, slot, func = self._get_pci_address_fields(pcie_addr)
        from nova.virt.libvirt.config import LibvirtConfigGuestHostdevPCI
        conf = LibvirtConfigGuestHostdevPCI()
        conf.domain = domain
        conf.bus = bus
        conf.slot = slot
        conf.function = func
        return conf

    def attach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        conf = self.get_config(instance, inst_type, acc)
        state = guest.get_power_state(host)
        from nova.compute import power_state
        live = state in (power_state.RUNNING, power_state.PAUSED)
        guest.attach_device(conf, persistent=True, live=live)

    def detach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        conf = self.get_config(instance, inst_type, acc)
        state = guest.get_power_state(host)
        from nova.compute import power_state
        live = state in (power_state.RUNNING, power_state.PAUSED)
        guest.detach_device(conf, persistent=True, live=live)

    def _get_pci_address_fields(self, pci_addr):
        """Parse a fully-specified PCI device address.

        :param pci_addr: A string of the form
        "<domain>:<bus>:<slot>.<function>".
        :return: A 4-tuple of strings
        ("<domain>", "<bus>", "<slot>", "<function>")
        """
        dbs, sep, func = pci_addr.partition('.')
        domain, bus, slot = dbs.split(':')
        return domain, bus, slot, func



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import abc
from oslo_config import cfg
import six


CONF = cfg.CONF


@six.add_metaclass(abc.ABCMeta)
class PluginBase(object):
    """Base class for all ACC plugins."""

    # Override to provide a tuple of oslo_config.Opt instances for
    # the plugin config parameters
    CONFIG_OPTS = ()

    def __init__(self, config):
        """
        Initialize the plugin object with the provided config

        :param config: `oslo_config.ConfigOpts.GroupAttr` instance:
        """
        self.config = config

    @abc.abstractmethod
    def get_config(self, acc, instance=None, inst_type=None):
        """
        Generate config of specified accelerator.

        :param instance: Instance information dict.
        :param inst_type: Flavor information dict.
        :param acc: Accelerator information
        :return: Different LibvirtConfig. For example,
        `nova.virt.libvirt.config.LibvirtConfigGuestHostdevPCI` object.
        """

    @abc.abstractmethod
    def attach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        """
        Given required parameters, perform operations to attach the ACC
        properly.

        :param acc: Accelerator information
        :param guest: `nova.virt.libvirt.guest.Guest` object.
        :param host: `nova.virt.libvirt.guest.Host` object.
        :param instance: Instance information dict.
        :param inst_type: Flavor information dict.
        :return:
        """

    @abc.abstractmethod
    def detach(self, acc, guest=None, host=None, instance=None,
               inst_type=None):
        """
        Given required parameters, perform operations to detach the ACC
        properly.

        :param acc: Accelerator information
        :param guest: `nova.virt.libvirt.guest.Guest` object.
        :param host: `nova.virt.libvirt.guest.Host` object.
        :param instance: Instance information dict.
        :param inst_type: Flavor information dict.
        :return:
        """

    @classmethod
    def load(cls, plugin_name):
        """
        Load a plugin, registering its configuration options

        :param plugin_name: the name of the plugin extension

        :returns: an initialized instance of the class
        """
        cfg_group_name = "os_acc_" + plugin_name
        cfg_opts = getattr(cls, "CONFIG_OPTS")
        cfg_vals = None
        if cfg_opts and len(cfg_opts) > 0:
            cfg_group = cfg.OptGroup(
                cfg_group_name,
                "os-acc plugin %s options" % plugin_name)
            CONF.register_opts(cfg_opts, group=cfg_group)

            cfg_vals = getattr(CONF, cfg_group_name)
        return cls(cfg_vals)



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from stevedore import extension

import os_acc.exception
from os_acc.i18n import _

_EXT_MANAGER = None
LOG = logging.getLogger(__name__)


def initialize(reset=False):
    global _EXT_MANAGER
    if reset or (_EXT_MANAGER is None):
        _EXT_MANAGER = extension.ExtensionManager(namespace='os_acc',
                                                  invoke_on_load=False)
        loaded_plugins = []
        for plugin_name in _EXT_MANAGER.names():
            cls = _EXT_MANAGER[plugin_name].plugin
            obj = cls.load(plugin_name)
            LOG.debug(("Loaded ACC plugin class '%(cls)s' "
                       "with name '%(plugin_name)s'"),
                      {'cls': cls, 'plugin_name': plugin_name})
            loaded_plugins.append(plugin_name)
            _EXT_MANAGER[plugin_name].obj = obj
        LOG.info("Loaded ACC plugins: %s", ", ".join(loaded_plugins))


def get_config(acc, instance=None, inst_type=None):

    if _EXT_MANAGER is None:
        raise os_acc.exception.LibraryNotInitialized()
    acc_type = acc['type']
    if acc_type is None:
        raise os_acc.exception.InternalError(
            _("acc_type parameter must be present "
              "for this acc_type implementation"))
    try:
        plugin = _EXT_MANAGER[acc_type].obj
    except KeyError:
        raise os_acc.exception.NoMatchingPlugin(plugin_name=acc_type)
    conf = plugin.get_config(acc, instance, inst_type)
    return conf


def attach(acc, guest=None, host=None, instance=None, inst_type=None):
    if _EXT_MANAGER is None:
        raise os_acc.exception.LibraryNotInitialized()
    acc_type = acc['type']
    if acc_type is None:
        raise os_acc.exception.InternalError(
            _("acc_type parameter must be present "
              "for this acc_type implementation"))
    try:
        plugin = _EXT_MANAGER[acc_type].obj
    except KeyError:
        raise os_acc.exception.NoMatchingPlugin(plugin_name=acc_type)
    try:
        LOG.debug("Attach acc %s", acc)
        plugin.attach(acc, guest, host, instance, inst_type)
        LOG.info("Successfully Attached acc %s", acc)
    except Exception as err:
        LOG.error("Failed to attach acc %(acc)s",
                  {"acc": acc}, exc_info=True)
        raise os_acc.exception.AttachException(acc=acc, err=err)


def detach(acc, guest=None, host=None, instance=None, inst_type=None):

    if _EXT_MANAGER is None:
        raise os_acc.exception.LibraryNotInitialized()
    acc_type = acc['type']
    if acc_type is None:
        raise os_acc.exception.InternalError(
            _("acc_type parameter must be present "
              "for this acc_type implementation"))
    try:
        plugin = _EXT_MANAGER[acc_type].obj
    except KeyError:
        raise os_acc.exception.NoMatchingPlugin(plugin_name=acc_type)

    try:
        LOG.debug("Detaching acc %s", acc)
        plugin.detach(acc, guest, host, instance, inst_type)
        LOG.info("Successfully detached acc %s", acc)
    except Exception as err:
        LOG.error("Failed to unplug acc %(acc)s",
                  {"acc": acc}, exc_info=True)
        raise os_acc.exception.DetachException(acc=acc, err=err)
